;==========================================================
; XREPACE.LSP 
; Copyright 1995 by Looking Glass Microproducts
;==========================================================
; Replace existing Xrefs
; 
;==========================================================
(Defun c:xreplace (/ error old_error notrans item bitset get_blocks
                   get_xref basename xreplace_main)

   ;=======================================================
   ; Error Handler
   ;=======================================================
   (Defun error (s)
      (If (Not
             (Member
                S
                '("Function cancelled" "console break")
             )
          )
         (Alert s)
      )
      (While (/= 0 (GetVar "cmdactive"))
         (Command)
      )
      (Command ".undo" "end")
      (Princ "\nUndoing...")
      (Command ".undo" 1)
      (Setq *Error* old_error)
      (Princ)
   )

   ;=======================================================
   ; Disallow transparent invocation of routine.
   ;=======================================================
   (Defun notrans ()
      (Cond
         ((Zerop (GetVar "cmdactive")))
         ((Alert
             "\nThis command may not be invoked transparently."
          )
         )
      )
   )

   ;=======================================================
   ; Return an item n from an association list.
   ;=======================================================
   (Defun item (n l) (CDR (Assoc n l)))

   ;=======================================================
   ; Return T if any bit of A matches B 
   ;=======================================================
   (Defun bitset (A B) (/= 0 (LogAnd A B)))

   ;=======================================================
   ; Get names of blocks in drawing
   ;=======================================================
   (Defun get_blocks (/ rewind bname flags path overlay entry blocks
                      xrefs)
      (Setq rewind T)
      (While (Setq entry (TblNext "block" rewind))
         (Setq
            rewind  Nil
            bname   (Strcase (item 2 entry))
            path    (item 1 entry)
            flags   (item 70 entry)
            overlay (bitset flags 8)
         )
         (Cond

            ;--------------------------------------
            ; ignore anonymous or dependent blocks
            ;--------------------------------------
            ((bitset flags (+ 1 16)))

            ;--------------------------------------
            ; note standard blocks
            ;--------------------------------------
            ((Not (bitset flags 4))
               (Setq blocks (Cons bname blocks))
            )

            ;--------------------------------------
            ; note xrefs
            ;--------------------------------------
            (T
               ;-----------------------------------
               ; Add extension to path if necessary
               ;-----------------------------------
               (Setq path (Strcase path T))
               (If (WCmatch path "~*`.*")
                  (Setq path (Strcat path ".dwg"))
               )
               (Setq
                  xrefs (Cons (List bname overlay path) xrefs)
               )
            )
         )
      )
      (List blocks xrefs)
   )
   ;==================================================
   ; Get name of existing Xref with a dialog box
   ;==================================================
   (Defun get_xref (xrefs / dcl_id choice choices r12)
      (Cond

         ;--------------------------------------------
         ; Load DCL File
         ;--------------------------------------------
         ((Minusp
             (Setq
                dcl_id (Load_Dialog "xreplace.dcl")
             )
          )
            (Alert
               "XREPLACE.DCL -- Can't load file."
            )
         )

         ;--------------------------------------------
         ; Start Dialog Box
         ;--------------------------------------------
         ((Not (New_Dialog "xreplace" dcl_id))
            Nil
         )
         (T
            (Setq choices (MapCar 'CAR xrefs))
            ;-----------------------------------------
            ; Sort the choices
            ;-----------------------------------------
            (If acad_strlsort
               (Setq choices (acad_strlsort choices))
            )

            ;-----------------------------------------
            ; Display choices in dialog box
            ;-----------------------------------------
            (Start_List "choices")
            (MapCar
               '(Lambda
                  (xref / temp path xtype)
                  (Setq
                     temp  (item xref xrefs)
                     xtype (If (CAR temp) "Overlay" "Attach")
                     path  (CADR temp)
                  )
                  (Add_List
                     (Strcat xref "\t" path "\t" xtype)
                  )
               )
               choices
            )
            (End_List)

            ;----------------------------------------
            ; Note the user choice
            ;----------------------------------------
            (Action_Tile
               "choices"
               "(setq choice (nth (read $value) choices))"
            )

            ;----------------------------------------
            ; Clear choice if cancelled
            ;----------------------------------------
            (If (Zerop (Start_Dialog))
               (Setq choice Nil)
            )
            choice
         )
      )
   )

   ;======================================================
   ; Strip drive, directory and extension from filename
   ;======================================================
   (Defun basename (fname)

      ;------------------------------------------
      ; Strip the drive and directory from fname
      ;------------------------------------------
      (While (WCmatch fname "*[:/\\]*")
         (Setq fname (Substr fname 2))
      )

      ;-----------------------------------------
      ; Strip the extension from fname
      ;-----------------------------------------
      (While (WCmatch fname "*`.*")
         (Setq
            fname (Substr fname 1 (1- (Strlen fname)))
         )
      )
      fname
   )
   ;======================================================
   ; XReplace Main
   ;======================================================
   (Defun xreplace_main (/ temp blocks xrefs oldname oldfname ext
                         flags newfname newname temp overlay)
      (Cond
         ;-------------------------------------------------
         ; Get the names of the xrefs and blocks in drawing
         ;-------------------------------------------------
         ((Progn
             (Setq
                temp   (get_blocks)
                blocks (CAR temp)
                xrefs  (CADR temp)
             )
             (Null xrefs)
          )
            (Alert
               "This drawing contains no Xrefs."
            )
         )

         ;-------------------------------------------------
         ; Get the name of the xref to replace
         ;-------------------------------------------------
         ((Null (Setq oldname (get_xref xrefs))))

         ;------------------------------------------------
         ; Get the new xref to attach
         ;------------------------------------------------
         ((Progn
             (Setq
                temp     (item oldname xrefs)
                overlay  (CAR temp)
                title    (Strcat
                            "Select File to "
                            (If overlay "Overlay" "Attach")
                         )
                oldfname (CADR temp)
                ext      "dwg"
                flags    (+ 8 2) ; library search, disable typeit
                newfname (GetFileD title oldfname ext flags)
             )
             (Null newfname)
          )
         )

         ;------------------------------------------------
         ; If the filename hasn't changed, reload the xref
         ;------------------------------------------------
         ((Progn
             (Setq
                newfname (Strcase newfname T)
                newname  (Strcase (basename newfname))
             )
             (= newname oldname)
          )
            (Princ
               (Strcat
                  "\n"
                  "Reloading Xref "
                  oldname
                  ": "
                  oldfname
                  "..."
               )
            )
            (Command
               ".xref" "path" oldname newfname
            )
            (Princ
               (Strcat "\nXref " oldname " reloaded.")
            )
         )

         ;------------------------------------------------
         ; Is the new xref already a standard block?
         ;------------------------------------------------
         ((Member newname blocks)
            (Alert
               (Strcat
                  newname
                  " is already a standard block"
                  " in the current drawing."
               )
            )
         )

         ;------------------------------------------------
         ; Is the new xref already in the drawing
         ;------------------------------------------------
         ((Assoc newname xrefs)
            (Alert
               (Strcat
                  newname
                  " is already an Xref "
                  "in the current drawing."
               )
            )
         )
         (T
            ;---------------------------------------------
            ; attach or overlay the new xref
            ;---------------------------------------------
            (Setq
               action (If overlay "Overlay" "Attach")
            )
            (Princ
               (Strcat
                  "\n"
                  action
                  "ing Xref "
                  newname
                  ": "
                  newfname
                  "..."
               )
            )
            (Command ".xref" action newfname)
            (Command)
            (Princ
               (Strcat "\nXref " newname " loaded.")
            )

            ;---------------------------------------------
            ; Find the insertions of the old xref
            ;---------------------------------------------
            (Setq
               ss (SsGet
                     "x"
                     (List
                        (Cons 0 "INSERT")
                        (Cons 2 oldname)
                     )
                  )
               n  (If ss (SsLength ss) 0)
            )

            ;---------------------------------------------
            ; Replace them with insertions of the new xref
            ;---------------------------------------------
            (If ss
               (Princ
                  (Strcat
                     "\nReplacing "
                     (ItoA n)
                     " insertion"
                     (If (= 1 n) "" "s")
                     " of "
                     oldname
                     " with "
                     newname
                     "..."
                  )
               )
            )
            (While (> n 0)
               (Setq n (1- n))
               (EntMod
                  (Subst
                     (Cons 2 newname)
                     (Cons 2 oldname)
                     (EntGet (SsName ss n))
                  )
               )
            )
            ;---------------------------------------------
            ; Detach the old xref
            ;---------------------------------------------
            (Command ".xref" "detach" oldname)
            (Princ
               (Strcat "\nXref " oldname " detached.")
            )

            ;---------------------------------------------
            ; And we're done!
            ;---------------------------------------------
            (If ss (Command ".redrawall"))
            (Setq ss Nil)
         )
      )
   )
   ;=======================================================
   ; Start of C:XREPLACE
   ;=======================================================
   (If (notrans)
      (Progn
         (Setq old_error *Error* *Error* error)
         (Setvar "cmdecho" 0)
         (Command ".undo" "group")
         (xreplace_main)
         (Command ".undo" "end")
         (Setq *Error* old_error)
      )
   )
   (Princ)
)
(Princ
   (Strcat
      "\nXREPLACE.LSP -- (c) 1995 "
      "by Looking Glass Microproducts"
   )
)
(Princ)
